from __future__ import annotations

import os
from dataclasses import dataclass

import pytest
from exasol_integration_test_docker_environment.lib.test_environment.ports import Ports

import exasol.pytest_backend.config as config


@dataclass
class OnpremDBConfig:
    """Exasol database configuration"""

    host: str
    port: int
    username: str
    password: str


@dataclass
class OnpremBfsConfig:
    """Bucketfs configuration"""

    url: str
    username: str
    password: str


@dataclass
class SshConfig:
    """SSH configuration"""

    port: int


@dataclass
class ItdeConfig:
    """Itde configuration settings"""

    db_version: str
    db_mem_size: str
    db_disk_size: str
    nameserver: list


_ONPREM_DB_OPTIONS = config.OptionGroup(
    prefix="exasol",
    options=(
        {
            "name": "host",
            "type": str,
            "default": "localhost",
            "help_text": "Host to connect to",
        },
        {
            "name": "port",
            "type": int,
            "default": Ports.forward.database,
            "help_text": "Port on which the exasol db is listening",
        },
        {
            "name": "username",
            "type": str,
            "default": "SYS",
            "help_text": "Username used to authenticate against the exasol db",
        },
        {
            "name": "password",
            "type": str,
            "default": "exasol",
            "help_text": "Password used to authenticate against the exasol db",
        },
    ),
)

_ONPREM_BFS_OPTIONS = config.OptionGroup(
    prefix="bucketfs",
    options=(
        {
            "name": "url",
            "type": str,
            "default": f"http://127.0.0.1:{Ports.forward.bucketfs}",
            "help_text": "Base url used to connect to the bucketfs service",
        },
        {
            "name": "username",
            "type": str,
            "default": "w",
            "help_text": "Username used to authenticate against the bucketfs service",
        },
        {
            "name": "password",
            "type": str,
            "default": "write",
            "help_text": "Password used to authenticate against the bucketfs service",
        },
    ),
)

_SSH_OPTIONS = config.OptionGroup(
    prefix="ssh",
    options=(
        {
            "name": "port",
            "type": int,
            "default": Ports.forward.ssh,
            "help_text": "Port on which external processes can access the database via SSH protocol",
        },
    ),
)

DEFAULT_ITDE_DB_VERSION = "8.18.1"
DEFAULT_ITDE_DB_MEM_SIZE = "2 GiB"
DEFAULT_ITDE_DB_DISK_SIZE = "2 GiB"

_ITDE_OPTIONS = config.OptionGroup(
    prefix="itde",
    options=(
        {
            "name": "db_version",
            "type": str,
            "default": DEFAULT_ITDE_DB_VERSION,
            "help_text": "DB version to start, if value is 'external' an existing instance will be used",
        },
        {
            "name": "db_mem_size",
            "type": str,
            "default": DEFAULT_ITDE_DB_MEM_SIZE,
            "help_text": (
                "The main memory used by the database. Format <number> <unit>, e.g. 1 GiB. "
                "The minimum size is 1 GB, below that the database will not start."
            ),
        },
        {
            "name": "db_disk_size",
            "type": str,
            "default": DEFAULT_ITDE_DB_DISK_SIZE,
            "help_text": (
                "The disk size available for the database. Format <number> <unit>, e.g. 1 GiB. "
                "The minimum size is 100 MiB. However, the setup creates volume files with "
                "at least 2 GB larger size, because the database needs at least so much more disk."
            ),
        },
        {
            "name": "nameserver",
            "type": str,
            "default": [],
            "help_text": (
                "A nameserver to be added to the list of DNS nameservers which the docker-db "
                "should use for resolving domain names. This option can be repeated "
                "multiple times."
            ),
        },
    ),
)

_ITDE_OPTION_GROUPS = (
    _ONPREM_DB_OPTIONS,
    _ONPREM_BFS_OPTIONS,
    _ITDE_OPTIONS,
    _SSH_OPTIONS,
)


def _add_option_group(parser, group):
    parser_group = parser.getgroup(group.prefix)
    for option in group.options:
        parser_group.addoption(
            option.cli,
            default=option.default,
            type=option.type,
            help=option.help,
            action="append" if isinstance(option.default, list) else "store",
        )


def itde_pytest_addoption(parser):
    for group in _ITDE_OPTION_GROUPS:
        _add_option_group(parser, group)


@pytest.fixture(scope="session")
def exasol_config(request) -> OnpremDBConfig:
    """Returns the configuration settings of the exasol db for this session."""
    cli_arguments = request.config.option
    kwargs = _ONPREM_DB_OPTIONS.kwargs(os.environ, cli_arguments)
    return OnpremDBConfig(**kwargs)


@pytest.fixture(scope="session")
def bucketfs_config(request) -> OnpremBfsConfig:
    """Returns the configuration settings of the bucketfs for this session."""
    cli_arguments = request.config.option
    kwargs = _ONPREM_BFS_OPTIONS.kwargs(os.environ, cli_arguments)
    return OnpremBfsConfig(**kwargs)


@pytest.fixture(scope="session")
def ssh_config(request) -> SshConfig:
    """Returns the configuration settings for SSH access in this session."""
    cli_arguments = request.config.option
    kwargs = _SSH_OPTIONS.kwargs(os.environ, cli_arguments)
    return SshConfig(**kwargs)


@pytest.fixture(scope="session")
def itde_config(request) -> ItdeConfig:
    """Returns the configuration settings of the ITDE for this session."""
    cli_arguments = request.config.option
    kwargs = _ITDE_OPTIONS.kwargs(os.environ, cli_arguments)
    return ItdeConfig(**kwargs)



"""ATTENTION:
This file is generated by exasol/toolbox/nox/_package_version.py when using:
   * either "poetry run -- nox -s project:fix"
   * or "poetry run -- nox -s version:check -- --fix"
Do not edit this file manually!
If you need to change the version, do so in the pyproject.toml, e.g. by using
`poetry version X.Y.Z`.
"""

MAJOR = 1
MINOR = 1
PATCH = 0
VERSION = f"{MAJOR}.{MINOR}.{PATCH}"
__version__ = VERSION



from __future__ import annotations

import multiprocessing as mp
from contextlib import (
    AbstractContextManager,
    ContextDecorator,
    contextmanager,
)
from functools import wraps
from typing import Any


class _ParallelGenCtxManager(AbstractContextManager, ContextDecorator):

    def __init__(self, func, args, kwargs):
        self._func_name = func.__name__
        self._ctx_func = contextmanager(func)
        self._args = args
        self._kwargs = kwargs
        self._queue = mp.Queue()
        self._ready = mp.Event()
        self._done = mp.Event()

    def _run(self) -> None:
        try:
            with self._ctx_func(*self._args, **self._kwargs) as output:
                # Now the task has completed. Make the results available for
                # caller and indicate the completion.
                # Return None as the error object followed by the function output.
                self._queue.put(None)
                self._queue.put(output)
                self._ready.set()
                # Wait until the caller is done and proceed with the cleanup.
                self._done.wait()
        except Exception as ex:
            # Pass the exception back to the caller and wait until it's done.
            self._queue.put(ex)
            self._ready.set()
            self._done.wait()

    def __enter__(self) -> _ParallelGenCtxManager:
        self._proc: mp.Process = mp.Process(target=self._run)
        self._proc.start()
        # Leave the process running
        return self

    def __exit__(self, *exc) -> None:
        # At this point the task should be completed or the process killed.
        # Indicate the task that it can now clean up and finish.
        self._done.set()
        self._proc.join()
        self._proc.close()

    def wait(self, timeout: float | None = None) -> None:
        if not self._ready.wait(timeout):
            self._proc.kill()
            raise TimeoutError(
                f"{self._func_name} failed to complete within {timeout} seconds"
            )

        # Check if an exception is returned and if so re-raise it
        error = self._queue.get()
        if error is not None:
            raise RuntimeError(f"{self._func_name} failed") from error

    def get_output(self, timeout: float | None = None) -> Any:
        self.wait(timeout)
        return self._queue.get()


def paralleltask(func):
    """
    @paralleltask decorator.

    The decorator starts a long-running function in a separate process. The function
    must be a generator. The targeted use case is a function that builds some kind of
    environment or a data structure, lets the caller use it, and then destroys it.
    Here is a rough pattern of such a function:

    def prepare_something(some_arguments):
        # do the preparation
        yield some_output
        # clean up

    Normally, this function would be used with the help of the @contextmanager decorator.

    @contextmanager
    def prepare_something(some_arguments):
        ...

    with prepare_something(some_arguments) as some_output:
        ...

    This decorator provides similar functionality but runs the function in a separate
    process. It changes the type of the returned output, wrapping it into an object with
    synchronisation functions. The output can be obtained using the get_output([timeout])
    method. This method blocks until the output is ready or the timeout expires. If the
    function produces no output the wait([timeout]) method can be called instead.

    The decorator can be used with session fixtures that want to start bringing up their
    resources in advance, possibly in parallel with each other. Below is a fixture that
    performs all its tasks at once, when it is called. Here we assume that prepare_something
    function is decorated with the @contextmanager.

    @fixture(score="session")
    def some_fixture():
        with prepare_something(...) as some_output:
            yield some_output

    This fixture can be split into two parts. Let's now assume that prepare_something is
    decorated with the @paralleltask.

    @fixture(score="session", autouse=True)
    def some_fixture_async():
        with prepare_something(...) as prep_task:
            yield prep_task

    @fixture(score="session")
    def some_fixture(some_fixture_async):
        return some_fixture_async.get_output()

    The first fixture will start bringing up the resources required for the second fixture.
    All fixtures of the first type should run before any of the fixtures of the second type,
    which is facilitated by the "autouse" parameter.
    """

    @wraps(func)
    def helper(*args, **kwargs):
        return _ParallelGenCtxManager(func, args, kwargs)

    return helper



"""
A "Project Short Tag" is a short abbreviation for a project.

The pytest plugin for exasol-saas-api will include this short tag into the
names of created database instances to enable identifying the origin of
potentially long-running database instances in order to avoid unwanted costs.
"""

from __future__ import annotations

from pathlib import Path

import yaml

YML_FILE = "error_code_config.yml"
STOP_FILE = "pyproject.toml"


def _find_path_backwards(
    start_path: str | Path,
    stop_file: str | Path = STOP_FILE,
    target_path: str | Path = YML_FILE,
) -> Path | None:
    """
    An utility searching for a specified path backwards. It begins with the given start
    path and checks if the target path is among its siblings. Then it moves to the parent
    path and so on, until it either reaches the root of the file structure or finds the
    stop file. returns None if the search is unsuccessful.
    """
    current_path = Path(start_path)
    root = Path(current_path.root)
    while current_path != root:
        result_path = Path(current_path, target_path)
        if result_path.exists():
            return result_path
        stop_path = Path(current_path, stop_file)
        if stop_path.exists():
            return None
        current_path = current_path.parent
    return None


def read_from_yaml(start_dir: Path) -> str | None:
    """
    Read project-short-tag from yaml file ``FILE`` looking for it from the
    specified starting directory ``start_dir``.
    If the yml file cannot be found returns None.
    If the yml file is found, but it doesn't have the expected format raises
    a RuntimeError.
    """
    config_file = _find_path_backwards(start_dir)
    if config_file is None:
        return None
    with config_file.open("r") as file:
        ecc = yaml.safe_load(file)
        try:
            return next(t for t in ecc["error-tags"])
        except Exception as ex:
            raise RuntimeError(
                f"Could not read project short tag from file {config_file}"
            )



from __future__ import annotations

import os
import ssl
from collections.abc import Generator
from datetime import timedelta
from typing import (
    Any,
)
from urllib.parse import urlparse

import pytest
from exasol.saas import client as saas_client
from exasol.saas.client.api_access import (
    OpenApiAccess,
    create_saas_client,
    get_connection_params,
    timestamp_name,
)
from exasol_integration_test_docker_environment.lib import api
from exasol_integration_test_docker_environment.lib.models.data.environment_info import (
    EnvironmentInfo,
)

import exasol.pytest_backend.project_short_tag as pst

from .itde import (
    bucketfs_config,
    exasol_config,
    itde_config,
    itde_pytest_addoption,
    ssh_config,
)
from .parallel_task import paralleltask

BACKEND_OPTION = "--backend"
BACKEND_ONPREM = "onprem"
BACKEND_SAAS = "saas"
BACKEND_ALL = "all"


def pytest_addoption(parser):
    parser.addoption(
        BACKEND_OPTION,
        action="append",
        default=[],
        help=f"""List of test backends ({BACKEND_ONPREM}, {BACKEND_SAAS}). The target 
            backends must be specified explicitly. To select all backends add the argument 
            {BACKEND_OPTION}={BACKEND_ALL} to the command line. To select only one
            of the backends add the argument {BACKEND_OPTION}={BACKEND_ONPREM} or
            {BACKEND_OPTION}={BACKEND_SAAS} instead.
            """,
    )
    parser.addoption(
        "--project-short-tag",
        help="""Short tag aka. "abbreviation" for your current project.
            See docstring in project_short_tag.py for more details.
            pytest plugin for exasol-saas-api will include this short tag into
            the names of created database instances.""",
    )
    parser.addoption(
        "--saas-database-id",
        help="""ID of the instance of an existing SaaS database to be
            used during the current pytest session instead of creating a
            dedicated instance temporarily.""",
    )
    parser.addoption(
        "--keep-saas-database",
        action="store_true",
        default=False,
        help="""Keep the SaaS database instance created for the current
            pytest session for subsequent inspection or reuse.""",
    )
    parser.addoption(
        "--saas-max-idle-hours",
        action="store",
        default=saas_client.Limits.AUTOSTOP_DEFAULT_IDLE_TIME.total_seconds() / 3600,
        help="""
        The SaaS cluster would normally stop after a certain period of inactivity. 
        The default period is 2 hours. For some tests, this period is too short.
        Use this parameter to set a sufficient idle period in the number of hours.
        """,
    )
    itde_pytest_addoption(parser)


def _is_backend_selected(request, backend: str) -> bool:
    backend_options = set(request.config.getoption(BACKEND_OPTION))
    return bool(backend_options.intersection({backend, BACKEND_ALL}))


@pytest.fixture(scope="session", params=[BACKEND_ONPREM, BACKEND_SAAS])
def backend(request) -> str:
    if not _is_backend_selected(request, request.param):
        pytest.skip()
    return request.param


@pytest.fixture(scope="session")
def use_onprem(request) -> bool:
    return _is_backend_selected(request, BACKEND_ONPREM)


@pytest.fixture(scope="session")
def use_saas(request) -> bool:
    return _is_backend_selected(request, BACKEND_SAAS)


@paralleltask
def start_itde(itde_config, exasol_config, bucketfs_config, ssh_config, database_name):
    """
    This function controls the ITDE with the help of parallelized
    version of the @contextmanager.
    """
    bucketfs_url = urlparse(bucketfs_config.url)
    env_info, cleanup_func = api.spawn_test_environment(
        environment_name=database_name,
        database_port_forward=exasol_config.port,
        bucketfs_port_forward=bucketfs_url.port,
        ssh_port_forward=ssh_config.port,
        db_mem_size=itde_config.db_mem_size,
        db_disk_size=itde_config.db_disk_size,
        nameserver=tuple(itde_config.nameserver),
        docker_db_image_version=itde_config.db_version,
    )
    yield env_info
    cleanup_func()


@pytest.fixture(scope="session", autouse=True)
def backend_aware_onprem_database_async(
    use_onprem, itde_config, exasol_config, bucketfs_config, ssh_config, database_name
):
    """
    If the onprem is a selected backend, this fixture starts brining up the ITDE and keeps
    it running for the duration of the session. It returns before the ITDE is ready. The
    "autouse" parameter assures that the ITDE, and other lengthy setup procedures are started
    preemptively before they are needed anywhere in the tests.

    The fixture shall not be used directly in tests.
    """
    if use_onprem and (itde_config.db_version != "external"):
        with start_itde(
            itde_config, exasol_config, bucketfs_config, ssh_config, database_name
        ) as itde_task:
            yield itde_task
    else:
        yield None


@pytest.fixture(scope="session")
def backend_aware_onprem_database(
    backend_aware_onprem_database_async,
) -> EnvironmentInfo | None:
    """
    If the onprem is a selected backend, this fixture waits till the ITDE becomes available.
    """
    if backend_aware_onprem_database_async is not None:
        return backend_aware_onprem_database_async.get_output()
    return None


def _env(var: str) -> str:
    result = os.environ.get(var)
    if result:
        return result
    raise RuntimeError(f"Environment variable {var} is empty.")


@pytest.fixture(scope="session")
def saas_host(use_saas) -> str:
    return _env("SAAS_HOST") if use_saas else ""


@pytest.fixture(scope="session")
def saas_pat(use_saas) -> str:
    return _env("SAAS_PAT") if use_saas else ""


@pytest.fixture(scope="session")
def saas_account_id(use_saas) -> str:
    return _env("SAAS_ACCOUNT_ID") if use_saas else ""


@pytest.fixture(scope="session")
def project_short_tag(request):
    return (
        request.config.getoption("--project-short-tag")
        or os.environ.get("PROJECT_SHORT_TAG")
        or pst.read_from_yaml(request.config.rootpath)
    )


@pytest.fixture(scope="session")
def database_name(project_short_tag):
    return timestamp_name(project_short_tag)


@pytest.fixture(scope="session")
def saas_api_access(
    request, use_saas, saas_host, saas_pat, saas_account_id
) -> Generator[OpenApiAccess | None]:
    if use_saas and (not request.config.getoption("--saas-database-id")):
        client = create_saas_client(host=saas_host, pat=saas_pat)
        api_access = OpenApiAccess(client=client, account_id=saas_account_id)
        with api_access.allowed_ip():
            yield api_access
    else:
        yield None


@pytest.fixture(scope="session", autouse=True)
def backend_aware_saas_database_id_async(
    request, use_saas, database_name, saas_api_access
) -> Generator[str]:
    """
    If the saas is a selected backend, this fixture starts building a temporary SaaS
    database and keeps it running for the duration of the session. It returns before the
    database is ready. The "autouse" parameter assures that the SaaS database, and other
    lengthy setup procedures are started preemptively before they are needed anywhere in
    the tests.

    The fixture shall not be used directly in tests.
    """
    if saas_api_access is not None:
        keep = request.config.getoption("--keep-saas-database")
        idle_hours = float(request.config.getoption("--saas-max-idle-hours"))

        # Create a temporary database
        with saas_api_access.database(
            name=database_name, keep=keep, idle_time=timedelta(hours=idle_hours)
        ) as db:
            yield db.id
    elif use_saas:
        yield request.config.getoption("--saas-database-id")
    else:
        yield ""


@pytest.fixture(scope="session")
def backend_aware_saas_database_id(
    saas_api_access, backend_aware_saas_database_id_async
) -> Generator[str]:
    """
    If the saas is a selected backend, this fixture waits until the temporary SaaS database
    becomes available.
    """
    if saas_api_access is not None:
        saas_api_access.wait_until_running(backend_aware_saas_database_id_async)
    yield backend_aware_saas_database_id_async


@pytest.fixture(scope="session")
def backend_aware_onprem_database_params(
    use_onprem, backend_aware_onprem_database, exasol_config
) -> dict[str, Any]:
    if use_onprem:
        return {
            "dsn": f"{exasol_config.host}:{exasol_config.port}",
            "user": exasol_config.username,
            "password": exasol_config.password,
            "websocket_sslopt": {"cert_reqs": ssl.CERT_NONE},
        }
    return {}


@pytest.fixture(scope="session")
def backend_aware_saas_database_params(
    use_saas, saas_host, saas_pat, saas_account_id, backend_aware_saas_database_id
) -> dict[str, Any]:
    if use_saas:
        conn_params = get_connection_params(
            host=saas_host,
            account_id=saas_account_id,
            database_id=backend_aware_saas_database_id,
            pat=saas_pat,
        )
        conn_params["websocket_sslopt"] = {"cert_reqs": ssl.CERT_NONE}
        return conn_params
    return {}


@pytest.fixture(scope="session")
def backend_aware_onprem_bucketfs_params(
    use_onprem, backend_aware_onprem_database, bucketfs_config
) -> dict[str, Any]:
    if use_onprem:
        return {
            "backend": BACKEND_ONPREM,
            "url": bucketfs_config.url,
            "username": bucketfs_config.username,
            "password": bucketfs_config.password,
            "service_name": "bfsdefault",
            "bucket_name": "default",
            "verify": False,
        }
    return {}


@pytest.fixture(scope="session")
def backend_aware_saas_bucketfs_params(
    use_saas, saas_host, saas_pat, saas_account_id, backend_aware_saas_database_id
) -> dict[str, Any]:
    if use_saas:
        return {
            "backend": BACKEND_SAAS,
            "url": saas_host,
            "account_id": saas_account_id,
            "database_id": backend_aware_saas_database_id,
            "pat": saas_pat,
        }
    return {}


@pytest.fixture(scope="session")
def backend_aware_database_params(
    backend, backend_aware_onprem_database_params, backend_aware_saas_database_params
) -> dict[str, Any]:
    """
    Returns a set of parameters sufficient to open a pyexasol connection to the
    current testing backend.
    Usage example:
    connection = pyexasol.connect(**backend_aware_database_params, compression=True)
    """
    if backend == BACKEND_ONPREM:
        return backend_aware_onprem_database_params
    elif backend == BACKEND_SAAS:
        return backend_aware_saas_database_params
    raise ValueError(f"Unknown backend {backend}")


@pytest.fixture(scope="session")
def backend_aware_bucketfs_params(
    backend, backend_aware_onprem_bucketfs_params, backend_aware_saas_bucketfs_params
) -> dict[str, Any]:
    """
    Returns a set of parameters sufficient to open a PathLike bucket-fs connection to the
    current testing backend.
    Usage example:
    bfs_path = exasol.bucketfs.path.build_path(**backend_aware_bucketfs_params, path=path_in_bucket)
    """
    if backend == BACKEND_ONPREM:
        return backend_aware_onprem_bucketfs_params
    elif backend == BACKEND_SAAS:
        return backend_aware_saas_bucketfs_params
    raise ValueError(f"Unknown backend {backend}")
